use anyhow::{anyhow, Result};
use clap::Parser;
use dropshot::{ConfigDropshot, ServerBuilder};

use slog::Drain;
use tracing_slog::TracingSlogDrain;

use std::net::SocketAddr;
use std::str::FromStr;
use std::sync::Arc;

use patagia_controller::api;
use patagia_controller::context::ControllerContext;

#[derive(Parser, Debug)]
#[command(version, about, long_about = None)]
struct Cli {
    #[arg(
        long = "telemetry-otlp-endpoint",
        default_value = "http://localhost:4317",
        value_name = "OTEL_EXPORTER_OTLP_ENDPOINT"
    )]
    otlp_endpoint: Option<String>,

    #[arg(
        long = "log-stderr",
        short = 'v',
        default_value = "false",
        value_name = "LOG_TO_STDERR"
    )]
    log_stderr: bool,
}

#[tokio::main]
async fn main() -> Result<()> {
    let args = Cli::parse();
    let _tracing = instrumentation::init_tracing(args.otlp_endpoint.as_ref(), args.log_stderr)?;

    tracing::info!("Patagia Controller");

    let config = ConfigDropshot {
        bind_address: SocketAddr::from_str("0.0.0.0:9474").unwrap(),
        ..Default::default()
    };

    // Adapt the Dropshot logger to tracing
    let logger = {
        let level_drain = slog::LevelFilter(TracingSlogDrain, slog::Level::Debug).fuse();
        let async_drain = slog_async::Async::new(level_drain).build().fuse();
        slog::Logger::root(async_drain, slog::o!())
    };

    let ctx = ControllerContext::new();
    let api = api::api()?;
    ServerBuilder::new(api, Arc::new(ctx), logger)
        .config(config)
        .start()
        .map_err(|e| anyhow!("Error starting server: {:?}", e))?
        .await
        .map_err(|e| anyhow!(e))
}
